@@ -9,17 +9,23 @@ different versions of the Mastodon API.
 """
 
 import dataclasses
+import typing as t
 
 from dataclasses import dataclass, is_dataclass
 from datetime import date, datetime
 from functools import lru_cache
-from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union
-from typing import get_type_hints
+from typing import Any, Dict, NamedTuple, Optional, Type, TypeVar, Union
+from typing import get_args, get_origin, get_type_hints
 
-from toot.typing_compat import get_args, get_origin
 from toot.utils import get_text
 from toot.utils.datetime import parse_datetime
 
+# Generic data class instance
+T = TypeVar("T")
+
+# A dict decoded from JSON
+Data = Dict[str, Any]
+
 
 @dataclass
 class AccountField:
@@ -59,8 +65,8 @@ class Account:
     header: str
     header_static: str
     locked: bool
-    fields: List[AccountField]
-    emojis: List[CustomEmoji]
+    fields: t.List[AccountField]
+    emojis: t.List[CustomEmoji]
     bot: bool
     group: bool
     discoverable: Optional[bool]
@@ -76,7 +82,7 @@ class Account:
     source: Optional[dict]
 
     @staticmethod
-    def __toot_prepare__(obj: Dict) -> Dict:
+    def __toot_prepare__(obj: Data) -> Data:
         # Pleroma has not yet converted last_status_at from datetime to date
         # so trim it here so it doesn't break when converting to date.
         # See: https://git.pleroma.social/pleroma/pleroma/-/issues/1470
@@ -154,10 +160,10 @@ class Poll:
     multiple: bool
     votes_count: int
     voters_count: Optional[int]
-    options: List[PollOption]
-    emojis: List[CustomEmoji]
+    options: t.List[PollOption]
+    emojis: t.List[CustomEmoji]
     voted: Optional[bool]
-    own_votes: Optional[List[int]]
+    own_votes: Optional[t.List[int]]
 
 
 @dataclass
@@ -207,11 +213,11 @@ class Filter:
     """
     id: str
     title: str
-    context: List[str]
+    context: t.List[str]
     expires_at: Optional[datetime]
     filter_action: str
-    keywords: List[FilterKeyword]
-    statuses: List[FilterStatus]
+    keywords: t.List[FilterKeyword]
+    statuses: t.List[FilterStatus]
 
 
 @dataclass
@@ -220,7 +226,7 @@ class FilterResult:
     https://docs.joinmastodon.org/entities/FilterResult/
     """
     filter: Filter
-    keyword_matches: Optional[List[str]]
+    keyword_matches: Optional[t.List[str]]
     status_matches: Optional[str]
 
 
@@ -237,11 +243,11 @@ class Status:
     visibility: str
     sensitive: bool
     spoiler_text: str
-    media_attachments: List[MediaAttachment]
+    media_attachments: t.List[MediaAttachment]
     application: Optional[Application]
-    mentions: List[StatusMention]
-    tags: List[StatusTag]
-    emojis: List[CustomEmoji]
+    mentions: t.List[StatusMention]
+    tags: t.List[StatusTag]
+    emojis: t.List[CustomEmoji]
     reblogs_count: int
     favourites_count: int
     replies_count: int
@@ -259,14 +265,14 @@ class Status:
     muted: Optional[bool]
     bookmarked: Optional[bool]
     pinned: Optional[bool]
-    filtered: Optional[List[FilterResult]]
+    filtered: Optional[t.List[FilterResult]]
 
     @property
     def original(self) -> "Status":
         return self.reblog or self
 
     @staticmethod
-    def __toot_prepare__(obj: Dict) -> Dict:
+    def __toot_prepare__(obj: Data) -> Data:
         # Pleroma has a bug where created_at is set to an empty string.
         # To avoid marking created_at as optional, which would require work
         # because we count on it always existing, set it to current datetime.
@@ -289,8 +295,8 @@ class Report:
     comment: str
     forwarded: bool
     created_at: datetime
-    status_ids: Optional[List[str]]
-    rule_ids: Optional[List[str]]
+    status_ids: Optional[t.List[str]]
+    rule_ids: Optional[t.List[str]]
     target_account: Account
 
 
@@ -328,7 +334,7 @@ class InstanceConfigurationStatuses:
 
 @dataclass
 class InstanceConfigurationMediaAttachments:
-    supported_mime_types: List[str]
+    supported_mime_types: t.List[str]
     image_size_limit: int
     image_matrix_limit: int
     video_size_limit: int
@@ -377,13 +383,13 @@ class Instance:
     urls: InstanceUrls
     stats: InstanceStats
     thumbnail: Optional[str]
-    languages: List[str]
+    languages: t.List[str]
     registrations: bool
     approval_required: bool
     invites_enabled: bool
     configuration: InstanceConfiguration
     contact_account: Optional[Account]
-    rules: List[Rule]
+    rules: t.List[Rule]
 
 
 @dataclass
@@ -397,7 +403,7 @@ class Relationship:
     following: bool
     showing_reblogs: bool
     notifying: bool
-    languages: List[str]
+    languages: t.List[str]
     followed_by: bool
     blocking: bool
     blocked_by: bool
@@ -428,7 +434,7 @@ class Tag:
     """
     name: str
     url: str
-    history: List[TagHistory]
+    history: t.List[TagHistory]
     following: Optional[bool]
 
 
@@ -445,26 +451,37 @@ class FeaturedTag:
     last_status_at: datetime
 
 
-# Generic data class instance
-T = TypeVar("T")
+@dataclass
+class List:
+    """
+    Represents a list of some users that the authenticated user follows.
+    https://docs.joinmastodon.org/entities/List/
+    """
+    id: str
+    title: str
+    # This is a required field on Mastodon, but not supported on Pleroma/Akkoma
+    # see: https://git.pleroma.social/pleroma/pleroma/-/issues/2918
+    replies_policy: Optional[str]
+
+# ------------------------------------------------------------------------------
+
+
+class Field(NamedTuple):
+    name: str
+    type: Any
+    default: Any
 
 
 class ConversionError(Exception):
     """Raised when conversion fails from JSON value to data class field."""
-    def __init__(
-        self,
-        data_class: Type,
-        field_name: str,
-        field_type: Type,
-        field_value: Optional[str]
-    ):
+    def __init__(self, data_class: type, field: Field, field_value: Optional[str]):
         super().__init__(
-            f"Failed converting field `{data_class.__name__}.{field_name}` "
-            + f"of type `{field_type.__name__}` from value {field_value!r}"
+            f"Failed converting field `{data_class.__name__}.{field.name}` "
+            + f"of type `{field.type.__name__}` from value {field_value!r}"
         )
 
 
-def from_dict(cls: Type[T], data: Dict) -> T:
+def from_dict(cls: Type[T], data: Data) -> T:
     """Convert a nested dict into an instance of `cls`."""
     # Apply __toot_prepare__ if it exists
     prepare = getattr(cls, '__toot_prepare__', None)
@@ -472,19 +489,19 @@ def from_dict(cls: Type[T], data: Dict) -> T:
         data = prepare(data)
 
     def _fields():
-        for name, type, default in get_fields(cls):
-            value = data.get(name, default)
-            converted = _convert_with_error_handling(cls, name, type, value)
-            yield name, converted
+        for field in _get_fields(cls):
+            value = data.get(field.name, field.default)
+            converted = _convert_with_error_handling(cls, field, value)
+            yield field.name, converted
 
     return cls(**dict(_fields()))
 
 
-@lru_cache(maxsize=100)
-def get_fields(cls: Type) -> List[Tuple[str, Type, Any]]:
+@lru_cache
+def _get_fields(cls: type) -> t.List[Field]:
     hints = get_type_hints(cls)
     return [
-        (
+        Field(
             field.name,
             _prune_optional(hints[field.name]),
             _get_default_value(field)
@@ -493,11 +510,11 @@ def get_fields(cls: Type) -> List[Tuple[str, Type, Any]]:
     ]
 
 
-def from_dict_list(cls: Type[T], data: List[Dict]) -> List[T]:
+def from_dict_list(cls: Type[T], data: t.List[Data]) -> t.List[T]:
     return [from_dict(cls, x) for x in data]
 
 
-def _get_default_value(field):
+def _get_default_value(field: dataclasses.Field):
     if field.default is not dataclasses.MISSING:
         return field.default
 
@@ -507,21 +524,16 @@ def _get_default_value(field):
     return None
 
 
-def _convert_with_error_handling(
-    data_class: Type,
-    field_name: str,
-    field_type: Type,
-    field_value: Optional[str]
-):
+def _convert_with_error_handling(data_class: type, field: Field, field_value: Any) -> Any:
     try:
-        return _convert(field_type, field_value)
+        return _convert(field.type, field_value)
     except ConversionError:
         raise
     except Exception:
-        raise ConversionError(data_class, field_name, field_type, field_value)
+        raise ConversionError(data_class, field, field_value)
 
 
-def _convert(field_type, value):
+def _convert(field_type: Any, value: Any) -> Any:
     if value is None:
         return None
 
@@ -544,7 +556,7 @@ def _convert(field_type, value):
     raise ValueError(f"Not implemented for type '{field_type}'")
 
 
-def _prune_optional(field_type: Type) -> Type:
+def _prune_optional(field_type: type) -> type:
     """For `Optional[<type>]` returns the encapsulated `<type>`."""
     if get_origin(field_type) == Union:
         args = get_args(field_type)
