@@ -2,20 +2,26 @@ import logging
 import subprocess
 import urwid
 
+
 from concurrent.futures import ThreadPoolExecutor
+from typing import NamedTuple, Optional
+from datetime import datetime, timezone
 
 from toot import api, config, __version__, settings
-from toot.console import get_default_visibility
-from toot.exceptions import ApiError
+from toot import App, User
+from toot.cli import get_default_visibility
+from toot.utils.datetime import parse_datetime
 
 from .compose import StatusComposer
 from .constants import PALETTE
 from .entities import Status
+from .images import TuiScreen, load_image
 from .overlays import ExceptionStackTrace, GotoMenu, Help, StatusSource, StatusLinks, StatusZoom
 from .overlays import StatusDeleteConfirmation, Account
 from .poll import Poll
 from .timeline import Timeline
-from .utils import get_max_toot_chars, parse_content_links, copy_to_clipboard
+from .utils import get_max_toot_chars, parse_content_links, copy_to_clipboard, LRUCache
+from .widgets import ModalBox, RoundedLineBox
 
 logger = logging.getLogger(__name__)
 
@@ -25,6 +31,17 @@ urwid.set_encoding('UTF-8')
 DEFAULT_MAX_TOOT_CHARS = 500
 
 
+class TuiOptions(NamedTuple):
+    colors: int
+    media_viewer: Optional[str]
+    always_show_sensitive: bool
+    relative_datetimes: bool
+    cache_size: int
+    default_visibility: Optional[str]
+    image_format: Optional[str]
+    show_display_names: bool
+
+
 class Header(urwid.WidgetWrap):
     def __init__(self, app, user):
         self.app = app
@@ -80,10 +97,12 @@ class TUI(urwid.Frame):
     screen: urwid.BaseScreen
 
     @staticmethod
-    def create(app, user, args):
+    def create(app: App, user: User, options: TuiOptions):
         """Factory method, sets up TUI and an event loop."""
-        screen = TUI.create_screen(args)
-        tui = TUI(app, user, screen, args)
+        screen = TuiScreen()
+        screen.set_terminal_properties(options.colors)
+
+        tui = TUI(app, user, screen, options)
 
         palette = PALETTE.copy()
         overrides = settings.get_setting("tui.palette", dict, {})
@@ -101,23 +120,11 @@ class TUI(urwid.Frame):
 
         return tui
 
-    @staticmethod
-    def create_screen(args):
-        screen = urwid.raw_display.Screen()
-
-        # Determine how many colors to use
-        default_colors = 1 if args.no_color else 16
-        colors = settings.get_setting("tui.colors", int, default_colors)
-        logger.debug(f"Setting colors to {colors}")
-        screen.set_terminal_properties(colors)
-
-        return screen
-
-    def __init__(self, app, user, screen, args):
+    def __init__(self, app, user, screen, options: TuiOptions):
         self.app = app
         self.user = user
-        self.args = args
         self.config = config.load_config()
+        self.options = options
 
         self.loop = None  # late init, set in `create`
         self.screen = screen
@@ -139,12 +146,18 @@ class TUI(urwid.Frame):
         self.can_translate = False
         self.account = None
         self.followed_accounts = []
-        self.media_viewer = settings.get_setting("tui.media_viewer", str)
+        self.preferences = {}
+
+        if self.options.cache_size:
+            self.cache_max = 1024 * 1024 * self.options.cache_size
+        else:
+            self.cache_max = 1024 * 1024 * 10  # default 10MB
 
         super().__init__(self.body, header=self.header, footer=self.footer)
 
     def run(self):
         self.loop.set_alarm_in(0, lambda *args: self.async_load_instance())
+        self.loop.set_alarm_in(0, lambda *args: self.async_load_preferences())
         self.loop.set_alarm_in(0, lambda *args: self.async_load_timeline(
             is_initial=True, timeline_name="home"))
         self.loop.set_alarm_in(0, lambda *args: self.async_load_followed_accounts())
@@ -175,8 +188,8 @@ class TUI(urwid.Frame):
 
         return urwid.Filler(intro)
 
-    def run_in_thread(self, fn, args=[], kwargs={}, done_callback=None, error_callback=None):
-        """Runs `fn(*args, **kwargs)` asynchronously in a separate thread.
+    def run_in_thread(self, fn, done_callback=None, error_callback=None):
+        """Runs `fn` asynchronously in a separate thread.
 
         On completion calls `done_callback` if `fn` exited cleanly, or
         `error_callback` if an exception was caught. Callback methods are
@@ -200,7 +213,10 @@ class TUI(urwid.Frame):
                 logger.exception(exception)
                 self.loop.set_alarm_in(0, lambda *args: _error_callback(exception))
 
-        future = self.executor.submit(fn, *args, **kwargs)
+        # TODO: replace by `self.loop.event_loop.run_in_executor` at some point
+        # Added in https://github.com/urwid/urwid/issues/575
+        # Not yet released at the time of this comment
+        future = self.loop.event_loop._loop.run_in_executor(self.executor, fn)
         future.add_done_callback(_done)
         return future
 
@@ -320,19 +336,34 @@ class TUI(urwid.Frame):
                 # get the major version number of the server
                 # this works for Mastodon and Pleroma version strings
                 # Mastodon versions < 4 do not have translation service
+                # If the version is missing, assume 0 as a fallback
                 # Revisit this logic if Pleroma implements translation
-                ch = instance["version"][0]
+                version = instance["version"]
+                ch = "0" if not version else version[0]
                 self.can_translate = int(ch) > 3 if ch.isnumeric() else False
 
         return self.run_in_thread(_load_instance, done_callback=_done)
 
+    def async_load_preferences(self):
+        """
+        Attempt to update user preferences from instance.
+        https://docs.joinmastodon.org/methods/preferences/
+        """
+        def _load_preferences():
+            return api.get_preferences(self.app, self.user).json()
+
+        def _done(preferences):
+            self.preferences = preferences
+
+        return self.run_in_thread(_load_preferences, done_callback=_done)
+
     def async_load_followed_accounts(self):
         def _load_accounts():
             try:
                 acct = f'@{self.user.username}@{self.user.instance}'
                 self.account = api.find_account(self.app, self.user, acct)
                 return api.following(self.app, self.user, self.account["id"])
-            except ApiError:
+            except Exception:
                 # not supported by all Mastodon servers so fail silently if necessary
                 return []
 
@@ -400,11 +431,45 @@ class TUI(urwid.Frame):
         def _post(timeline, *args):
             self.post_status(*args)
 
-        composer = StatusComposer(self.max_toot_chars, self.user.username, in_reply_to)
+        # If the user specified --default-visibility, use that; otherwise,
+        # try to use the server-side default visibility.  If that fails, fall
+        # back to get_default_visibility().
+        visibility = (self.options.default_visibility or
+                      self.preferences.get('posting:default:visibility',
+                                           get_default_visibility()))
+
+        composer = StatusComposer(self.max_toot_chars, self.user.username,
+                                  visibility, in_reply_to)
         urwid.connect_signal(composer, "close", _close)
         urwid.connect_signal(composer, "post", _post)
         self.open_overlay(composer, title="Compose status")
 
+    def async_edit(self, status):
+        def _fetch_source():
+            return api.fetch_status_source(self.app, self.user, status.id).json()
+
+        def _done(source):
+            self.close_overlay()
+            self.show_edit(status, source)
+
+        please_wait = ModalBox("Loading status...")
+        self.open_overlay(please_wait)
+
+        self.run_in_thread(_fetch_source, done_callback=_done)
+
+    def show_edit(self, status, source):
+        def _close(*args):
+            self.close_overlay()
+
+        def _edit(timeline, *args):
+            self.edit_status(status, *args)
+
+        composer = StatusComposer(self.max_toot_chars, self.user.username,
+                                  visibility=None, edit=status, source=source)
+        urwid.connect_signal(composer, "close", _close)
+        urwid.connect_signal(composer, "post", _edit)
+        self.open_overlay(composer, title="Edit status")
+
     def show_goto_menu(self):
         user_timelines = self.config.get("timelines", {})
         user_lists = api.get_lists(self.app, self.user) or []
@@ -503,8 +568,15 @@ class TUI(urwid.Frame):
         if not urls:
             return
 
-        if self.media_viewer:
-            subprocess.run([self.media_viewer] + urls)
+        media_viewer = self.options.media_viewer
+        if media_viewer:
+            try:
+                subprocess.run([media_viewer] + urls)
+            except FileNotFoundError:
+                self.footer.set_error_message(f"Media viewer not found: '{media_viewer}'")
+            except Exception as ex:
+                self.exception = ex
+                self.footer.set_error_message("Failed invoking media viewer. Press X to see exception.")
         else:
             self.footer.set_error_message("Media viewer not configured")
 
@@ -545,11 +617,47 @@ class TUI(urwid.Frame):
         self.footer.set_message("Status posted {} \\o/".format(status.id))
         self.close_overlay()
 
+    def edit_status(self, status, content, warning, visibility, in_reply_to_id):
+        # We don't support editing polls (yet), so to avoid losing the poll
+        # data from the original toot, copy it to the edit request.
+        poll_args = {}
+        poll = status.original.data.get('poll', None)
+
+        if poll is not None:
+            poll_args['poll_options'] = [o['title'] for o in poll['options']]
+            poll_args['poll_multiple'] = poll['multiple']
+
+            # Convert absolute expiry time into seconds from now.
+            expires_at = parse_datetime(poll['expires_at'])
+            expires_in = int((expires_at - datetime.now(timezone.utc)).total_seconds())
+            poll_args['poll_expires_in'] = expires_in
+
+            if 'hide_totals' in poll:
+                poll_args['poll_hide_totals'] = poll['hide_totals']
+
+        data = api.edit_status(
+            self.app,
+            self.user,
+            status.id,
+            content,
+            spoiler_text=warning,
+            visibility=visibility,
+            **poll_args
+        ).json()
+
+        new_status = self.make_status(data)
+
+        self.footer.set_message("Status edited {} \\o/".format(status.id))
+        self.close_overlay()
+
+        if self.timeline is not None:
+            self.timeline.update_status(new_status)
+
     def show_account(self, account_id):
         account = api.whois(self.app, self.user, account_id)
         relationship = api.get_relationship(self.app, self.user, account_id)
         self.open_overlay(
-            widget=Account(self.app, self.user, account, relationship),
+            widget=Account(self.app, self.user, account, relationship, self.options),
             title="Account",
         )
 
@@ -658,6 +766,27 @@ class TUI(urwid.Frame):
 
         return self.run_in_thread(_delete, done_callback=_done)
 
+    def async_load_image(self, timeline, status, path, placeholder_index):
+        def _load():
+            # don't bother loading images for statuses we are not viewing now
+            if timeline.get_focused_status().id != status.id:
+                return
+
+            if not hasattr(timeline, "images"):
+                timeline.images = LRUCache(cache_max_bytes=self.cache_max)
+
+            img = load_image(path)
+            if img:
+                timeline.images[str(hash(path))] = img
+
+        def _done(loop):
+            # don't bother loading images for statuses we are not viewing now
+            if timeline.get_focused_status().id != status.id:
+                return
+            timeline.update_status_image(status, path, placeholder_index)
+
+        return self.run_in_thread(_load, done_callback=_done)
+
     def copy_status(self, status):
         # TODO: copy a better version of status content
         # including URLs
@@ -672,7 +801,7 @@ class TUI(urwid.Frame):
     )
 
     def open_overlay(self, widget, options={}, title=""):
-        top_widget = urwid.LineBox(widget, title=title)
+        top_widget = RoundedLineBox(widget, title=title)
         bottom_widget = self.body
 
         _options = self.default_overlay_options.copy()
